/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#ifndef _ERFF4_H_
#define _ERFF4_H_	1

#include <altivec.h>
#include <vec_types.h>

#include "expf4.h"
#include "recipf4.h"
#include "divf4.h"

#define SQRT_PI          1.772453850905516027298167483341E0f
#define INV_SQRT_PI      5.641895835477562869480794515608E-1f
#define TWO_OVER_SQRT_PI 1.128379167095512573896158903122E0f

/*
 * Coefficients of Taylor Series Expansion of Error Function
 */
#define TAYLOR_ERF_00  1.00000000000000000000000000000000E0f
#define TAYLOR_ERF_01 -3.33333333333333333333333333333333E-1f
#define TAYLOR_ERF_02  1.00000000000000000000000000000000E-1f
#define TAYLOR_ERF_03 -2.38095238095238095238095238095238E-2f
#define TAYLOR_ERF_04  4.62962962962962962962962962962963E-3f
#define TAYLOR_ERF_05 -7.57575757575757575757575757575758E-4f
#define TAYLOR_ERF_06  1.06837606837606837606837606837607E-4f
#define TAYLOR_ERF_07 -1.32275132275132275132275132275132E-5f
#define TAYLOR_ERF_08  1.45891690009337068160597572362278E-6f
#define TAYLOR_ERF_09 -1.45038522231504687645038522231505E-7f
#define TAYLOR_ERF_10  1.31225329638028050726463424876123E-8f
#define TAYLOR_ERF_11 -1.08922210371485733804574384284529E-9f
#define TAYLOR_ERF_12  8.35070279514723959168403612848057E-11f
#define TAYLOR_ERF_13 -5.94779401363763503681199154450183E-12f
#define TAYLOR_ERF_14  3.95542951645852576339713723402831E-13f
#define TAYLOR_ERF_15 -2.46682701026445692771004257606067E-14f
#define TAYLOR_ERF_16  1.44832646435981372649642651245986E-15f
#define TAYLOR_ERF_17 -8.03273501241577360913984452288663E-17f
#define TAYLOR_ERF_18  4.22140728880708823303144982433982E-18f
#define TAYLOR_ERF_19 -2.10785519144213582486050800945443E-19f
#define TAYLOR_ERF_20  1.00251649349077191670194893132589E-20f
#define TAYLOR_ERF_21 -4.55184675892820028624362194732684E-22f
#define TAYLOR_ERF_22  1.97706475387790517483308832055610E-23f
#define TAYLOR_ERF_23 -8.23014929921422135684449347132513E-25f
#define TAYLOR_ERF_24  3.28926034917575173275247613224729E-26f
#define TAYLOR_ERF_25 -1.26410789889891635219506925866759E-27f
#define TAYLOR_ERF_26  4.67848351551848577372630857707162E-29f
#define TAYLOR_ERF_27 -1.66976179341737202698649397026798E-30f
#define TAYLOR_ERF_28  5.75419164398217177219656443388090E-32f
#define TAYLOR_ERF_29 -1.91694286210978253077267196219294E-33f
#define TAYLOR_ERF_30  6.18030758822279613746380577974771E-35f
#define TAYLOR_ERF_31 -1.93035720881510785655551537411475E-36f
#define TAYLOR_ERF_32  5.84675500746883629629795521967448E-38f
#define TAYLOR_ERF_33 -1.71885606280178362396819126765645E-39f
#define TAYLOR_ERF_34  4.90892396452342296700208077293189E-41f
#define TAYLOR_ERF_35 -1.36304126177913957635067836351026E-42f
#define TAYLOR_ERF_36  3.68249351546114573519399405666776E-44f
#define TAYLOR_ERF_37 -9.68728023887076175384366004096384E-46f
#define TAYLOR_ERF_38  2.48306909745491159103989919026756E-47f
#define TAYLOR_ERF_39 -6.20565791963739670594197460728991E-49f
#define TAYLOR_ERF_40  1.51310794954121709805375306782686E-50f
#define TAYLOR_ERF_41 -3.60157930981012591661339989697254E-52f
#define TAYLOR_ERF_42  8.37341968387228154282667202937594E-54f
#define TAYLOR_ERF_43 -1.90254122728987952723942026863661E-55f
#define TAYLOR_ERF_44  4.22678975419355257583834431489747E-57f
#define TAYLOR_ERF_45 -9.18642950239868569596123672834859E-59f

  /*
   * Taylor Series Expansion of Erf
   *
   *                       infinite
   *                      ---------
   *                       -            n     2n
   *            2 * x        -        -1  *  x
   * erf(x) =    ----    *    -      ------------
   *            sqrt(pi)     -        (2n + 1) * n!
   *                       -     
   *                      ---------
   *                       n = 0
   *
   */

#define TAYLOR_ERFF4(_xabs, _xsqu, _tresult)  {   \
  _tresult = vec_madd(_xsqu, (vector float) {TAYLOR_ERF_45, TAYLOR_ERF_45, TAYLOR_ERF_45, TAYLOR_ERF_45}, \
                             (vector float) {TAYLOR_ERF_44, TAYLOR_ERF_44, TAYLOR_ERF_44, TAYLOR_ERF_44}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_43, TAYLOR_ERF_43, TAYLOR_ERF_43, TAYLOR_ERF_43}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_42, TAYLOR_ERF_42, TAYLOR_ERF_42, TAYLOR_ERF_42}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_41, TAYLOR_ERF_41, TAYLOR_ERF_41, TAYLOR_ERF_41}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_40, TAYLOR_ERF_40, TAYLOR_ERF_40, TAYLOR_ERF_40}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_39, TAYLOR_ERF_39, TAYLOR_ERF_39, TAYLOR_ERF_39}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_38, TAYLOR_ERF_38, TAYLOR_ERF_38, TAYLOR_ERF_38}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_37, TAYLOR_ERF_37, TAYLOR_ERF_37, TAYLOR_ERF_37}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_36, TAYLOR_ERF_36, TAYLOR_ERF_36, TAYLOR_ERF_36}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_35, TAYLOR_ERF_35, TAYLOR_ERF_35, TAYLOR_ERF_35}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_34, TAYLOR_ERF_34, TAYLOR_ERF_34, TAYLOR_ERF_34}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_33, TAYLOR_ERF_33, TAYLOR_ERF_33, TAYLOR_ERF_33}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_32, TAYLOR_ERF_32, TAYLOR_ERF_32, TAYLOR_ERF_32}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_31, TAYLOR_ERF_31, TAYLOR_ERF_31, TAYLOR_ERF_31}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_30, TAYLOR_ERF_30, TAYLOR_ERF_30, TAYLOR_ERF_30}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_29, TAYLOR_ERF_29, TAYLOR_ERF_29, TAYLOR_ERF_29}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_28, TAYLOR_ERF_28, TAYLOR_ERF_28, TAYLOR_ERF_28}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_27, TAYLOR_ERF_27, TAYLOR_ERF_27, TAYLOR_ERF_27}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_26, TAYLOR_ERF_26, TAYLOR_ERF_26, TAYLOR_ERF_26}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_25, TAYLOR_ERF_25, TAYLOR_ERF_25, TAYLOR_ERF_25}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_24, TAYLOR_ERF_24, TAYLOR_ERF_24, TAYLOR_ERF_24}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_23, TAYLOR_ERF_23, TAYLOR_ERF_23, TAYLOR_ERF_23}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_22, TAYLOR_ERF_22, TAYLOR_ERF_22, TAYLOR_ERF_22}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_21, TAYLOR_ERF_21, TAYLOR_ERF_21, TAYLOR_ERF_21}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_20, TAYLOR_ERF_20, TAYLOR_ERF_20, TAYLOR_ERF_20}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_19, TAYLOR_ERF_19, TAYLOR_ERF_19, TAYLOR_ERF_19}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_18, TAYLOR_ERF_18, TAYLOR_ERF_18, TAYLOR_ERF_18}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_17, TAYLOR_ERF_17, TAYLOR_ERF_17, TAYLOR_ERF_17}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_16, TAYLOR_ERF_16, TAYLOR_ERF_16, TAYLOR_ERF_16}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_15, TAYLOR_ERF_15, TAYLOR_ERF_15, TAYLOR_ERF_15}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_14, TAYLOR_ERF_14, TAYLOR_ERF_14, TAYLOR_ERF_14}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_13, TAYLOR_ERF_13, TAYLOR_ERF_13, TAYLOR_ERF_13}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_12, TAYLOR_ERF_12, TAYLOR_ERF_12, TAYLOR_ERF_12}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_11, TAYLOR_ERF_11, TAYLOR_ERF_11, TAYLOR_ERF_11}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_10, TAYLOR_ERF_10, TAYLOR_ERF_10, TAYLOR_ERF_10}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_09, TAYLOR_ERF_09, TAYLOR_ERF_09, TAYLOR_ERF_09}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_08, TAYLOR_ERF_08, TAYLOR_ERF_08, TAYLOR_ERF_08}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_07, TAYLOR_ERF_07, TAYLOR_ERF_07, TAYLOR_ERF_07}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_06, TAYLOR_ERF_06, TAYLOR_ERF_06, TAYLOR_ERF_06}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_05, TAYLOR_ERF_05, TAYLOR_ERF_05, TAYLOR_ERF_05}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_04, TAYLOR_ERF_04, TAYLOR_ERF_04, TAYLOR_ERF_04}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_03, TAYLOR_ERF_03, TAYLOR_ERF_03, TAYLOR_ERF_03}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_02, TAYLOR_ERF_02, TAYLOR_ERF_02, TAYLOR_ERF_02}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_01, TAYLOR_ERF_01, TAYLOR_ERF_01, TAYLOR_ERF_01}); \
  _tresult = vec_madd(_tresult, _xsqu, (vector float) {TAYLOR_ERF_00, TAYLOR_ERF_00, TAYLOR_ERF_00, TAYLOR_ERF_00}); \
  _tresult = vec_madd(_tresult, _xabs, zerof);                                                 \
  _tresult = vec_madd(_tresult, (vector float){TWO_OVER_SQRT_PI, TWO_OVER_SQRT_PI, TWO_OVER_SQRT_PI, TWO_OVER_SQRT_PI}, zerof); \
}



  /*
   * Continued Fractions Approximation of Erfc()
   *                                             (                      )
   *                          1                 (  1    v   2v   3v      )
   *   erfc(x)  =  -------------------------  * ( ---  ---  ---  --- ... )
   *               sqrt(pi) * x * exp(x^2)      (  1+   1+  1+   1+      )
   *                                             (                      )
   *                                                Continued Fractions
   *          1
   *   v =  -----
   *        2*x^2
   *
   *   We are using a backward recurrence calculation to estimate the continued fraction.
   *
   *   p    =   a   p        +  b   q
   *     m,n     m   m+1,n       m    m+1,n
   *
   *   q    =   p
   *     m,n     m+1,n
   *
   *   With,
   *
   *   p    =   a   ;   q    =  1
   *    n,n      n        n,n
   *
   *
   *   a  =  0,    b   =  1,
   *    0           0
   *
   *   a  =  1,    b   =  n/2x^2
   *    n           n
   *
   *
   *    F     =   p    /    q
   *     0,n       0,n       0,n
   *
   * Ref: "Computing the Incomplete Gamma Function to Arbitrary Precision",
   *       by Serge Winitzki, Department of Physics, Ludwig-Maximilians University, Munich, Germany.
   *
   */

#define CONTFRAC_ERFCF4(_xabs, _xsqu, _presult) {   \
  vec_float4 v;                         \
  vec_float4 p, q, plast, qlast;        \
  vec_float4 factor;                    \
  vec_float4 inv_xsqu;                  \
  inv_xsqu = _recipf4(_xsqu);            \
  v = vec_madd(inv_xsqu, onehalff, zerof); \
  p = (vector float) {3.025f, 3.025f, 3.025f, 3.025f}; q = onef; plast = p; qlast = q;                                           \
  p = vec_madd(qlast, vec_madd(v, (vector float){40.0f, 40.0f, 40.0f, 40.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){39.0f, 39.0f, 39.0f, 39.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){38.0f, 38.0f, 38.0f, 38.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){37.0f, 37.0f, 37.0f, 37.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){36.0f, 36.0f, 36.0f, 36.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){35.0f, 35.0f, 35.0f, 35.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){34.0f, 34.0f, 34.0f, 34.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){33.0f, 33.0f, 33.0f, 33.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){32.0f, 32.0f, 32.0f, 32.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){31.0f, 31.0f, 31.0f, 31.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){30.0f, 30.0f, 30.0f, 30.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){29.0f, 29.0f, 29.0f, 29.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){28.0f, 28.0f, 28.0f, 28.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){27.0f, 27.0f, 27.0f, 27.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){26.0f, 26.0f, 26.0f, 26.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){25.0f, 25.0f, 25.0f, 25.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){24.0f, 24.0f, 24.0f, 24.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){23.0f, 23.0f, 23.0f, 23.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){22.0f, 22.0f, 22.0f, 22.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){21.0f, 21.0f, 21.0f, 21.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){20.0f, 20.0f, 20.0f, 20.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){19.0f, 19.0f, 19.0f, 19.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){18.0f, 18.0f, 18.0f, 18.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){17.0f, 17.0f, 17.0f, 17.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){16.0f, 16.0f, 16.0f, 16.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){15.0f, 15.0f, 15.0f, 15.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){14.0f, 14.0f, 14.0f, 14.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){13.0f, 13.0f, 13.0f, 13.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){12.0f, 12.0f, 12.0f, 12.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){11.0f, 11.0f, 11.0f, 11.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){10.0f, 10.0f, 10.0f, 10.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){ 9.0f,  9.0f,  9.0f,  9.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){ 8.0f,  8.0f,  8.0f,  8.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){ 7.0f,  7.0f,  7.0f,  7.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){ 6.0f,  6.0f,  6.0f,  6.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){ 5.0f,  5.0f,  5.0f,  5.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){ 4.0f,  4.0f,  4.0f,  4.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){ 3.0f,  3.0f,  3.0f,  3.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){ 2.0f,  2.0f,  2.0f,  2.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = vec_madd(qlast, vec_madd(v, (vector float){ 1.0f,  1.0f,  1.0f,  1.0f}, zerof), plast); q = plast; plast = p; qlast = q;    \
  p = qlast; q = plast;                                                                         \
  factor = vec_madd((vector float){SQRT_PI, SQRT_PI, SQRT_PI, SQRT_PI}, vec_madd(_xabs, _expf4(_xsqu), zerof), zerof); \
  _presult = _divf4(p, vec_madd(factor, q, zerof));                                                     \
}

/*
 * FUNCTION
 *  vector float _erff4(vector float x)
 *
 * DESCRIPTION
 *  The erff4 function computes the error function of each element of x.
 *
 *  C99 Special Cases:
 *  - erf(+0) returns +0
 *  - erf(-0) returns -0
 *  - erf(+infinite) returns +1
 *  - erf(-infinite) returns -1
 *
 */

static __inline vector float _erff4(vector float x)
{
  vector float onehalff  = (vector float) {0.5f, 0.5f, 0.5f, 0.5f};
  vector float zerof     = (vector float) {0.0f, 0.0f, 0.0f, 0.0f};
  vector float onef      = (vector float) {1.0f, 1.0f, 1.0f, 1.0f};
  vector float sign_mask = (vector float) {-0.0f, -0.0f, -0.0f, -0.0f};

  /* This is where we switch from Taylor Series to Continued Fraction approximation */
  vec_float4 approx_point = (vector float) {0.89f};

  vector bool isinf;
  vec_float4 xabs, xsqu, xsign;
  vec_float4 tresult, presult, result;

  xsign = vec_and(x, sign_mask);

  /* Force Denorms to 0 */
  x = vec_add(x, zerof);

  xabs = vec_andc(x, sign_mask);
  xsqu = vec_madd(x, x, zerof);

  /*
   * Taylor Series Expansion near Zero
   */
  TAYLOR_ERFF4(xabs, xsqu, tresult);

  /*
   * Continued Fraction Approximation of Erfc().
   * erf = 1 - erfc 
   */
  CONTFRAC_ERFCF4(xabs, xsqu, presult);
  presult = vec_sub(onef, presult);

  /*
   * Select the appropriate approximation.
   */
  result = vec_sel(tresult, presult, vec_cmpgt(xabs, approx_point));

  /*
   * Special cases/errors.
   */

  /* x = +/- infinite, return +/-1 */
  isinf = vec_cmpeq((vec_uint4)xabs, (vec_uint4){0x7F800000, 0x7F800000, 0x7F800000, 0x7F800000});
  result = vec_sel(result, onef, isinf);

  /*
   * Preserve sign in result, since erf(-x) = -erf(x)
   */
  result = vec_or(result, xsign);

  return result;
}

#endif /* _ERFF4_H_ */

